"""
    线程锁装饰器，直接锁整个函数
"""
import threading
from typing import Callable


class FuncLockDecorator:
    """
        独享锁-线程锁
    """

    def __init__(self, timeout: int = -1):
        self.timeout = timeout
        self.lock = threading.RLock()  # 包住整个函数的，所以在实例化里面

    def __call__(self, func: Callable):
        def _wrapper(*args, **kwargs):

            self.lock.acquire(timeout=self.timeout)

            try:
                return func(*args, **kwargs)
            finally:
                self.lock.release()

        return _wrapper


class FuncLockSharedDecorator:
    """
        共享锁-线程锁（用户不要使用该锁）

        一个进程内共享
    """
    LOCK = threading.RLock()  # 包住整个函数的，所以在实例化里面

    def __init__(self, timeout: int = 5):
        self.timeout = timeout

    def __call__(self, func: Callable):
        def _wrapper(*args, **kwargs):

            self.__class__.LOCK.acquire(timeout=self.timeout)

            try:
                return func(*args, **kwargs)
            finally:
                try:
                    self.__class__.LOCK.release()
                except:
                    pass

        return _wrapper



"""
    请求记录中间件

    目前记录：
        all：请求数
        succeed：请求成功数
        failed：请求失败数

    注意：
        通过引入 spider 自身的 spider_record 获取、修改对应信息
"""
import json
import datetime
from typing import Union
from palp import settings
from palp.network.request import Request
from palp.middleware.middleware_request import RequestMiddleware
from palp.decorator.decorator_lock import FuncLockSharedDecorator


class RequestRecordMiddleware(RequestMiddleware):
    """
        请求记录中间件
    """

    @FuncLockSharedDecorator(timeout=1)
    def request_in(self, spider, request) -> None:
        """
            记录进入的请求
        """
        spider.spider_record['all'] += 1

    @FuncLockSharedDecorator(timeout=1)
    def request_failed(self, spider, request) -> None:
        """
            记录失败的请求
        """
        spider.spider_record['failed'] += 1

    @FuncLockSharedDecorator(timeout=1)
    def request_close(self, spider, request, response) -> Union[Request, None]:
        """
            记录成功的请求
        """
        spider.spider_record['succeed'] += 1

        return

    def request_record(self, spider, record: dict) -> None:
        """
        发送总结果记录

        :param spider:
        :param record:
        :return:
        """
        from palp.conn import redis_conn

        if not redis_conn or settings.SPIDER_TYPE == 1:
            return

        record = {
            'all': record['all'],
            'failed': record['failed'],
            'succeed': record['succeed'],
            'stop_time': str(datetime.datetime.now())
        }

        redis_conn.set(settings.REDIS_KEY_STOP, json.dumps(record, ensure_ascii=False))



"""
    适用于 requests 的响应解析器
"""
from palp.network.response import Response


class RequestsResponse(Response):
    """
        适用于 requests 的响应解析器
    """

    def __init__(self, resp):
        super().__init__(resp)

    @property
    def text(self) -> str:
        return self.content.decode(self.encoding) if self.content is not None else ''

    @property
    def content(self) -> bytes:
        return self.response.content or b''

    def json(self, **kwargs) -> dict:
        return self.response.json(**kwargs)

    @property
    def status_code(self) -> int:
        return self.response.status_code

    @property
    def cookies(self) -> dict:
        return self.response.cookies.get_dict()

    @property
    def headers(self) -> dict:
        return dict(self.response.headers)

    @property
    def url(self) -> str:
        return self.response.url

    @property
    def history(self) -> list:
        """
        响应 url

        :return:
        """
        return self.response.history

    @property
    def links(self):
        """

        :return:
        """
        return self.response.links



"""
    分布式 spider
"""
import json
import time
import dill
import importlib
from palp import settings
from quickdb import RedisLock
from abc import abstractmethod
from palp.spider.spider import Spider
from palp.tool.client_heart import ClientHeart
from palp.tool.short_module import import_module
from palp.network.request import LoadRequest, Request
from palp.sequence.sequence_redis_item import FIFOItemRedisSequence
from palp.decorator.decorator_spider_wait import SpiderWaitDecorator
from palp.decorator.decorator_spider_record import SpiderRecordDecorator
from palp.decorator.decorator_run_func_by_thread import RunByThreadDecorator
from palp.decorator.decorator_spider_middleware import SpiderMiddlewareDecorator


class DistributiveSpider(Spider):
    """
        分布式 spider
    """

    def __new__(cls, *args, **kwargs):
        """
        加载 spider 设置，并定义 SPIDER_TYPE

        :param args:
        :param kwargs:
        """
        if not cls.SPIDER_MIDDLEWARE:
            setattr(settings, 'SPIDER_TYPE', 2)
            cls.from_settings()

        return object.__new__(cls)

    def __init__(self, thread_count=None, redis_key: str = None, request_filter=False, item_filter=False):
        """

        :param thread_count: 线程数量
        :param redis_key: 自定义 redis_key 的名字
        :param request_filter: 开启请求过滤，不然 filter_repeat 无效
        :param item_filter: 开启 item 过滤
        """
        self.spider_master = False
        self.redis_key = redis_key or self.name

        # 根据 spider 的名字设置 redis 前缀
        for key, value in settings.__dict__.items():
            if key.startswith('REDIS_KEY'):
                setattr(settings, key, value.format(redis_key=self.redis_key))

        super().__init__(thread_count, request_filter, item_filter)

        queue_module = settings.REQUEST_QUEUE[settings.SPIDER_TYPE][settings.REQUEST_QUEUE_MODE]
        self.queue = import_module(queue_module)[0]  # 请求队列
        self.queue_item = FIFOItemRedisSequence()  # item 队列

    @abstractmethod
    def start_requests(self) -> None:
        """
        起始函数

        :return:
        """
        pass

    @staticmethod
    def start_check():
        """
        启动检查，防止上次意外结束，导致 master 死机 key 未删除，无法正常启动
        这里设定时间 10s 即出意外后，超过 10s 启动才会移除 master key

        :return:
        """
        from palp.conn import redis_conn

        master_detail = redis_conn.get(settings.REDIS_KEY_MASTER)
        if master_detail:
            master_detail = json.loads(master_detail.decode())
            heart_beat = redis_conn.hget(settings.REDIS_KEY_HEARTBEAT, master_detail['name'])

            # 没有心跳的情况下，master 创建时间不能与现在时间相差 5s 不然就有问题
            if not heart_beat:
                if time.time() - master_detail['time'] > 5:
                    redis_conn.delete(settings.REDIS_KEY_MASTER)

            # 有心跳的情况下，master 超过 10s 没有跳动，就是有问题
            elif time.time() - json.loads(heart_beat.decode())['time'] > 10:
                redis_conn.delete(settings.REDIS_KEY_MASTER)

    @RunByThreadDecorator(daemon=True)
    def start_distribute_failed_request(self):
        """
        重新放入 执行失败的请求任务

        :return:
        """
        from palp.conn import redis_conn

        if settings.REQUEST_RETRY_FAILED and redis_conn.exists(settings.REDIS_KEY_QUEUE_BAD_REQUEST):
            while True:
                request = redis_conn.spop(settings.REDIS_KEY_QUEUE_BAD_REQUEST, count=1)
                if not request:
                    break

                self.queue.put(LoadRequest.load_from_json(request[0].decode()))

    @RunByThreadDecorator(daemon=True)
    def start_distribute_failed_item(self) -> None:
        """
        重新放入 处理失败的 item

        :return:
        """
        from palp.conn import redis_conn

        modules = {}
        if settings.ITEM_RETRY_FAILED and redis_conn.exists(settings.REDIS_KEY_QUEUE_BAD_ITEM):
            while True:
                item = redis_conn.spop(settings.REDIS_KEY_QUEUE_BAD_ITEM, count=1)
                if not item:
                    break

                item = json.loads(item[0].decode())

                # 重新导入 item
                key = item['module'] + '.' + item['init']
                if key not in modules:
                    cls = importlib.import_module(item['module']).__getattribute__(item['init'])  # 导入
                    modules[key] = cls
                else:
                    cls = modules[key]

                self.queue_item.put(cls(**json.loads(item['data'])))

    def competition_for_master(self) -> None:
        """
        竞争为 master 并执行任务分配

        注意：任务分配为线程分配，不然数量一旦过多，将会影响程序执行

        :return:
        """
        from palp.conn import redis_conn

        with RedisLock(conn=redis_conn, lock_name=settings.REDIS_KEY_LOCK + 'Master', block_timeout=10):
            if redis_conn.exists(settings.REDIS_KEY_MASTER):
                return

            # 设置 master 标志
            self.spider_master = True
            redis_conn.set(settings.REDIS_KEY_MASTER, json.dumps({
                'name': self.spider_uuid,
                'time': time.time(),
            }, ensure_ascii=False))

            # 删除之前的锁
            keys = redis_conn.keys(settings.REDIS_KEY_LOCK + '*')
            if keys:
                remove_list = []

                for key in keys:
                    key = key.decode()
                    if key.endswith('Master'):
                        continue
                    remove_list.append(key)

                if remove_list:
                    redis_conn.delete(*remove_list)

            # 删除 cookie 池
            if settings.REQUEST_BORROW_DELETE_WHEN_START:
                redis_conn.delete(settings.REDIS_KEY_QUEUE_REQUEST_BORROW)

            # 删除停止标志、记录
            redis_conn.delete(settings.REDIS_KEY_STOP, settings.REDIS_KEY_RECORD)

            # 清空心跳
            redis_conn.delete(settings.REDIS_KEY_HEARTBEAT, settings.REDIS_KEY_HEARTBEAT_FAILED)

            # 分发失败的任务
            self.start_distribute_failed_request()
            self.start_distribute_failed_item()

            # 分发任务
            self.start_distribute()

    @staticmethod
    def borrow_request(request: Request):
        """
        开启 REQUEST_BORROW 时，用来主动使用资源

        :param request: 新的请求，决定何时调用
        :return:
        """
        from palp import conn

        if settings.REQUEST_BORROW and conn.redis_conn:

            # 空 cookie_jar 即视为新的请求，添加 cookie_jar
            if not request.cookie_jar and request.keep_cookie:
                recycle_data = conn.redis_conn.lpop(settings.REDIS_KEY_QUEUE_REQUEST_BORROW)
                if recycle_data:
                    recycle_data = dill.loads(recycle_data)
                    if 'cookie_jar' in recycle_data:
                        request.cookie_jar = recycle_data['cookie_jar']

    @staticmethod
    def recycle_request(request: Request):
        """
        开启 REQUEST_BORROW 时，用来主动回收资源，一般在程序执行最后一个函数后调用

        尽量是 request 含有的 key

        :param request: 旧的请求。回收指定资源
        :return:
        """
        from palp import conn

        if settings.REQUEST_BORROW and conn.redis_conn:
            recycle_data = {}

            # 回收 cookie
            if request.cookie_jar:
                recycle_data.update({'cookie_jar': request.cookie_jar})

            # 发送回收的资源
            if recycle_data:
                conn.redis_conn.rpush(settings.REDIS_KEY_QUEUE_REQUEST_BORROW, dill.dumps(recycle_data))

    @SpiderMiddlewareDecorator()
    @SpiderRecordDecorator()
    @SpiderWaitDecorator()
    def run(self) -> None:
        """
        分布式处理 逻辑

        :return:
        """
        from palp.conn import redis_conn

        self.start_controller()  # 任务处理
        self.start_check()  # 检查是否正常
        self.competition_for_master()  # 竞争为 master

        # 等待任务分发完毕，客户端无任务，视为执行结束
        ClientHeart(spider=self).start()

        # master 机器处理的事情
        if self.spider_master:
            # 删除 master 的标志
            redis_conn.delete(settings.REDIS_KEY_MASTER)
            # 删除 heart 标志
            redis_conn.delete(settings.REDIS_KEY_HEARTBEAT, settings.REDIS_KEY_HEARTBEAT_FAILED)
            # 判断是否移除 filter
            if not settings.PERSISTENCE_REQUEST_FILTER:
                redis_conn.delete(settings.REDIS_KEY_QUEUE_FILTER_REQUEST)
            if not settings.PERSISTENCE_ITEM_FILTER:
                redis_conn.delete(settings.REDIS_KEY_QUEUE_FILTER_ITEM)



from palp.spider import JumpSpider
from palp.spider.spider_cycle import CycleSpider
from palp.tool.alarm import send_email
from palp.tool.alarm import send_dingtalk
from palp.tool import start_spider
from palp.item.item import Item
from palp.item.item_strict import StrictItem
from palp.item.item_lazy import LazyItem
from palp.network.request import Request
from palp.network.response import Response
from palp.pipeline.pipeline import Pipeline
from palp.spider.spider_local import LocalSpider
from palp.spider.spider_distributive import DistributiveSpider
from palp.network.request_method import RequestGet, RequestPost
from palp.middleware.middleware_spider import SpiderMiddleware
from palp.middleware.middleware_request import RequestMiddleware
from palp.exception import DropItemException, DropRequestException
from palp.item.item import Field



"""
    spider 资源回收中间件
"""
from palp import conn
from palp.middleware.middleware_spider import SpiderMiddleware


class SpiderRecycleMiddleware(SpiderMiddleware):
    """
        spider 资源回收中间件
    """

    def spider_close(self, spider):
        """
        关闭所有创建的连接

        :param spider:
        :return:
        """
        # 关闭 redis
        try:
            if conn.redis_conn_lazy:
                conn.redis_conn_lazy.close_all()
        except:
            pass

        # 关闭 pg 连接
        try:
            if conn.pg_conn:
                conn.pg_conn.close_all()
        except:
            pass

        # 关闭 mysql 连接
        try:
            if conn.mysql_conn:
                conn.mysql_conn.close_all()
        except:
            pass

        # 关闭 mongo 连接
        try:
            if conn.mongo_conn:
                conn.mongo_conn.close()
        except:
            pass

        # 关闭 kafka 连接
        try:
            if conn.kafka_conn:
                conn.kafka_conn.close()
        except:
            pass



"""
    请求中间件
"""
from typing import Union
from palp.network.request import Request
from palp.network.response import Response
from palp.spider.spider import Spider


class RequestMiddlewareBase:
    """
        请求中间件基类
    """

    def request_in(self, spider: Spider, request: Request) -> None:
        """
        请求进入时的操作

        :param spider:
        :param request:
        :return:
        """

    def request_error(self, spider: Spider, request: Request, exception: Exception) -> Union[Request, None]:
        """
        请求出错时的操作

        :param spider:
        :param request: 该参数可返回（用于放弃当前请求，并发起新请求）
        :param exception: 错误
        :return: [Request, None]
        """

    def request_failed(self, spider: Spider, request: Request) -> None:
        """
        超过最大重试次数时的操作

        :param spider:
        :param request:
        :return:
        """

    def request_close(self, spider: Spider, request: Request, response: Response) -> Union[Request, None]:
        """
        请求结束时的操作

        :param spider:
        :param request: 该参数可返回（用于放弃当前请求，并发起新请求）
        :param response:
        :return: [Request, None]
        """

    def request_record(self, spider: Spider, record: dict) -> None:
        """
        发送请求记录，在 spider 执行完毕时，将会调用

        :param spider:
        :param record:
        :return:
        """


class RequestMiddleware(RequestMiddlewareBase):
    """
        外部引用：请求中间件
    """

    def request_in(self, spider, request) -> None:
        """
        请求进入时的操作

        :param spider:
        :param request:
        :return:
        """

    def request_error(self, spider, request, exception: Exception) -> Union[Request, None]:
        """
        请求出错时的操作

        :param spider:
        :param request: 该参数可返回（用于放弃当前请求，并发起新请求）
        :param exception: 错误
        :return: [Request, None]
        """

    def request_failed(self, spider, request) -> None:
        """
        超过最大重试次数时的操作

        :param spider:
        :param request:
        :return:
        """

    def request_close(self, spider, request, response) -> Union[Request, None]:
        """
        请求结束时的操作

        :param spider:
        :param request: 该参数可返回（用于放弃当前请求，并发起新请求）
        :param response:
        :return: [Request, None]
        """

    def request_record(self, spider, record: dict) -> None:
        """
        发送请求记录，在 spider 执行完毕时，将会调用

        :param spider:
        :param record:
        :return:
        """



"""
    消息队列
"""
from typing import Any
from abc import abstractmethod


class SequenceBase:
    """
        队列的基类
    """

    @abstractmethod
    def put(self, obj: Any, timeout: float = None) -> None:
        """
        添加任务

        :param timeout:
        :param obj
        :return:
        """

    @abstractmethod
    def get(self, timeout: float = None) -> None:
        """
        获取任务

        :return:
        """

    @abstractmethod
    def empty(self) -> bool:
        """
        判断队列是否为空

        :return:
        """

    @abstractmethod
    def qsize(self) -> int:
        """
        返回队列大小

        :return:
        """


class Sequence(SequenceBase):
    """
        用于外部引用：队列的基类
    """

    def put(self, obj, timeout=None):
        """
        添加任务

        :param timeout:
        :param obj
        :return:
        """

    def get(self, timeout=None):
        """
        获取任务

        :return:
        """

    def empty(self):
        """
        判断队列是否为空

        :return:
        """

    def qsize(self):
        """
        返回队列大小

        :return:
        """


class RedisSequence(Sequence):
    """
        Redis 队列的基类
    """

    def __new__(cls, *args, **kwargs):
        """
        给创建的每一个类都添加一个 redis_key 属性

        :param args:
        :param kwargs:
        """
        cls.redis_key = cls.get_redis_key()

        return object.__new__(cls)

    @classmethod
    @abstractmethod
    def get_redis_key(cls) -> str:
        """
        获取 redis 的键

        :return:
        """



"""
    周期爬虫记录请求中间件

    注意：
        双继承的中间件，需要导入 2 次
        双继承的中间件，文件名、类名不含有 middleware_spider、middleware_request
"""
import json
from palp import settings
from quickdb import RedisLock
from palp.spider.spider import Spider
from palp.spider.spider_cycle import CycleSpider
from palp.middleware.middleware_spider import SpiderMiddleware
from palp.middleware.middleware_request import RequestMiddleware


class CycleSpiderRecordMiddleware(RequestMiddleware, SpiderMiddleware):
    """
        周期爬虫记录请求中间件
    """

    def spider_start(self, spider) -> None:
        """
        周期爬虫，启动插入初始记录

        :param spider:
        :return:
        """
        from palp.conn import redis_conn

        # 判断是否是 CycleSpider 并继承了其它 spider（没写一行是为了避免导入后被其它爬虫误用）
        if not isinstance(spider, CycleSpider):
            return
        if not issubclass(spider.__class__, Spider):
            return

        if settings.SPIDER_TYPE == 1:
            spider.insert_task_record_start()
        else:
            with RedisLock(conn=redis_conn, lock_name=settings.REDIS_KEY_LOCK + 'CycleSpiderRecord'):
                if self.check_max_id_is_done(spider=spider):
                    spider.insert_task_record_start()

    def request_failed(self, spider, request):
        """
        失败请求记录

        :param spider:
        :param request:
        :return:
        """
        if not isinstance(spider, CycleSpider):
            return
        if not issubclass(spider.__class__, Spider):
            return

        if hasattr(request, 'task_id'):
            spider.set_task_state_failed(task_id=request.task_id)

    def request_close(self, spider, request, response):
        """
        成功请求记录

        :param spider:
        :param request:
        :return:
        """
        if not isinstance(spider, CycleSpider):
            return
        if not issubclass(spider.__class__, Spider):
            return

        if hasattr(request, 'task_id'):
            spider.set_task_state_done(task_id=request.task_id)

    def request_record(self, spider, record: dict) -> None:
        """
        记录爬取结果

        :param spider:
        :param record:
        :return:
        """
        if not isinstance(spider, CycleSpider):
            return
        if not issubclass(spider.__class__, Spider):
            return

        spider.update_task_record(
            total=record['all'],
            succeed=record['succeed'],
            failed=record['failed'],
        )
        spider.update_task_record_end()

    @classmethod
    def check_max_id_is_done(cls, spider: CycleSpider) -> bool:
        """
        检查最大 id 的执行状态

        :param spider: spider
        :return:
        """
        from palp.conn import mysql_conn

        sql = f'''
            SELECT
                max( id ),
                is_done
            FROM
                `{spider.spider_table_record_name}`
        '''

        res = mysql_conn.execute(sql=sql, fetchone=True)

        if res[-1] is None:
            return True
        else:
            return bool(res[-1])



"""
    redis 上才布隆过滤器

    学习链接：https://cuiqingcai.com/8472.html
"""
from palp import settings
from quickdb import RedisLock
from palp.network.request import Request
from palp.filter.filter import FilterBase


class RedisBloomFilter(FilterBase):
    def __init__(self):
        """
        Initialize BloomFilter

        """
        self.m = 1 << settings.BLOOMFILTER_BIT
        self.seeds = range(settings.BLOOMFILTER_HASH_NUMBER)

    def is_repeat(self, obj, **kwargs):
        """
        获取对应的指纹，通过本地实现 redis 的 布隆过滤器 去重

        :param obj:
        :param kwargs:
        :return:
        """
        from palp.conn import redis_conn

        fingerprint = self.fingerprint(obj)

        if isinstance(obj, Request):
            redis_key_filter = settings.REDIS_KEY_QUEUE_FILTER_REQUEST
        else:
            redis_key_filter = settings.REDIS_KEY_QUEUE_FILTER_ITEM

        if settings.STRICT_FILTER:
            with RedisLock(conn=redis_conn, lock_name=settings.REDIS_KEY_LOCK + 'Request'):
                return self.judge(redis_key_filter, fingerprint)
        else:
            return self.judge(redis_key_filter, fingerprint)

    def exists(self, fingerprint: str, redis_key_filter: str):
        """
        判断值是否存在

        :return:
        """
        from palp.conn import redis_conn

        if not fingerprint:
            return False
        exist = 1
        for seed in self.seeds:
            offset = self.hash(seed, fingerprint)
            exist = exist & redis_conn.getbit(redis_key_filter, offset)
        return exist

    def insert(self, fingerprint: str, redis_key_filter: str):
        """
        把值加入到 bloom

        :return:
        """
        from palp.conn import redis_conn

        for seed in self.seeds:
            offset = self.hash(seed, fingerprint)
            redis_conn.setbit(redis_key_filter, offset, 1)

    def hash(self, seed: int, value: str):
        """
        Hash

        :return:
        """
        ret = 0
        for i in range(len(value)):
            ret += seed * ret + ord(value[i])
        return (self.m - 1) & ret

    def judge(self, f, fingerprint):
        """
        进行判断

        :param f: 判断条件或方法之类
        :param fingerprint: 指纹
        :return:
        """
        if self.exists(fingerprint=fingerprint, redis_key_filter=f):
            return True

        self.insert(fingerprint=fingerprint, redis_key_filter=f)

        return False
